﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Timers;

namespace Wsbis.CustomControls
{
    /// <summary>
    /// Consider the case of an object with PropertyChanged event.  It is possible that a single change may cause several properties to change, firing multiple
    /// PropertyChanged events.  If a listener is listening for PropertyChanged, and it does some expensive behavior, a naive implementation will cause that
    /// behavior to run multiple times.  This class addresses the issue by accumulating property change notifications for an object and only dispatching them
    /// (in a single event) after a certain delay has passed.
    /// </summary>
    public static class AccumulatedPropertyChangeNotification
    {        
        /// <summary>
        /// When a property change notification occurs on an object, we will wait at least this long (and maybe longer)
        /// before we call the accumulated handler. 
        /// </summary>
        public const int NOTIFICATION_DELAY_MS = 50;

        public delegate void AccumulatedPropertyChangedHandler(object sender, AccumulatedPropertyChangedEventArgs e);

        private static Timer timer;

        static AccumulatedPropertyChangeNotification()
        {
            timer = new Timer
            {
                AutoReset = true,
                Interval = NOTIFICATION_DELAY_MS                
            };
            timer.Elapsed += timer_Elapsed;            
            timer.Start();
        }

        private static string[] ExtractNotifications(IEnumerable<PropertyChangedEventArgs> changes)
        {
            // In general, if any are null or "", then we return null
            // otherwise, return distinct strings
            if (changes.Any(chg => String.IsNullOrEmpty(chg.PropertyName)))
                return null;

            return changes.Select(chg => chg.PropertyName).Distinct().ToArray();
        }        

        private static void timer_Elapsed(object sender, ElapsedEventArgs e)
        {
            // this list will be those change notifications processed and can be removed from queue
            Tuple<INotifyPropertyChanged, AccumulatedPropertyChangedHandler, ISynchronizeInvoke, string[]>[] changesToDispatch;

            lock (timer)
            {
                
                // find those objects which have change notifications, and when the notification was entered
                // we use group by here to group by the inotifypropertychanged objects and select the most recent datetime for each
                var latestReq = changeNotificationsQueued.GroupBy(cnq => cnq.Item1, (key, g) => new { INPC = key, Latest = g.Max(scnq => scnq.Item3), Items = g });
                

                // now we want to filter this resultset down to only those whose most recent datetime is in the past at least NOTIFICATION_DELAY_MS ago
                var delay = TimeSpan.FromMilliseconds(NOTIFICATION_DELAY_MS);
                var suitablyOldReq = latestReq.Where(req => DateTime.Now -  req.Latest >= delay).ToArray();
                

                // now suitablyOldReq will hold those objects which should be notified
                // for each object, we should prepare to remove from queue and invoke
                var changesServiced = suitablyOldReq.SelectMany(req => req.Items).ToArray();
                

                // Remove from the notification queue those notifications that will be handled
                foreach (var toRemove in changesServiced)
                    changeNotificationsQueued.Remove(toRemove);

                // extract into usable array.  For each INotifyPropertyChanged, extract the list of handlers (APCH)
                // and from each PropertyChangedEventArgs extract property name(s)
                // flatten as appropriate
                var toNotify = suitablyOldReq
                    .Select(req => new
                    {
                        APCH = notifications.Where(notf => notf.Item1 == req.INPC),                        
                        Changes = ExtractNotifications(req.Items.Select(it => it.Item2))
                    })
                    .SelectMany(req => req.APCH.Select(ap => new Tuple<INotifyPropertyChanged, AccumulatedPropertyChangedHandler, ISynchronizeInvoke, string[]>(ap.Item1, ap.Item2, ap.Item3, req.Changes)));

                // now this is done and we can leave the critical area
                changesToDispatch = toNotify.ToArray();
                

            }
            // we now leave the critical area, as the notification dispatch may occur onto other threads, etc, and does not concern the handler
            // we have "frozen" the changes to service into an array.

            foreach (var dispatch in changesToDispatch)
            {
                /*
                 * Tuple dispatch:
                 * 
                 * INotifyPropertyChanged,                  The object which originally generated the change notification
                 * AccumulatedPropertyChangedHandler,       The handler which will handle the change notification
                 * ISynchronizeInvoke,                      The synchronizer for the method invocation (may be null)
                 * string[]                                 One or more properties changed, or null if "all"
                 * */
                var ea = new AccumulatedPropertyChangedEventArgs { Properties = dispatch.Item4};
                // Check if the ISynchronizeInvoke is provided.  If so, use it to invoke the method.  Otherwise, invoke the method directly
                if (dispatch.Item3 != null && dispatch.Item3.InvokeRequired)
                {
                    dispatch.Item3.BeginInvoke(dispatch.Item2, new object[] { dispatch.Item1, ea });
                }
                else
                {
                    dispatch.Item2(dispatch.Item1, ea);
                }
            }
            
            
        }

        /// <summary>
        /// List of notifications for an object.  If a property changes on the object (Item1, left), then we should notify all handlers (Item2, right)
        /// </summary>
        private static List<Tuple<INotifyPropertyChanged, AccumulatedPropertyChangedHandler, ISynchronizeInvoke>> notifications = new List<Tuple<INotifyPropertyChanged, AccumulatedPropertyChangedHandler, ISynchronizeInvoke>>();

        /// <summary>
        /// When this class receives a "property changed" event notification, it is added to this queue.  Then the timer will watch for how old the events are.
        /// </summary>
        private static List<Tuple<INotifyPropertyChanged, PropertyChangedEventArgs, DateTime>> changeNotificationsQueued = new List<Tuple<INotifyPropertyChanged, PropertyChangedEventArgs, DateTime>>();

        /// <summary>
        /// Do we have any references to this object in our notification list?
        /// </summary>        
        private static bool AreWatchingObject(INotifyPropertyChanged objectToWatch)
        {
            return notifications.Any(n => Object.ReferenceEquals(n.Item1, objectToWatch));
        }

        /// <summary>
        /// Add a delayed PropertyChanged handler.  When the objectToWatch raises a PropertyChanged event, the event will be enqueued and bulk dispatched via the given
        /// handler when at least NOTIFICATION_DELAY_MS has elapsed.  If multiple PropertyChanged events occur from the same source, the delay will continue until
        /// no more events occur, then they will be dispatched together in a single event on handler.  If given, the handler call will be synchronized with
        /// a synchronizer object.
        /// </summary>
        /// <param name="objectToWatch">An object which may provide PropertyChanged events</param>
        /// <param name="handler">An event handler to handle the object's PropertyChanged events, implementing AccumulatedPropertyChangedHandler to handle events in batches</param>
        /// <param name="syncObject">Optional.  A synchronizing object for the handler.</param>
        public static void AddHandler(INotifyPropertyChanged objectToWatch, AccumulatedPropertyChangedHandler handler, ISynchronizeInvoke syncObject = null)
        {
            lock (timer)
            {
                // if we do not yet listen for notifications on this object, add a handler to the object's property change notifier
                if (!AreWatchingObject(objectToWatch))
                {
                    objectToWatch.PropertyChanged += objectToWatch_PropertyChanged;
                }

                notifications.Add(
                    new Tuple<INotifyPropertyChanged, AccumulatedPropertyChangedHandler, ISynchronizeInvoke>(
                        objectToWatch, handler, syncObject));
            }
            
        }

        /// <summary>
        /// Remove a previously added delayed PropertyChanged handler.  The objectToWatch and handler must be reference equals to the previously added handler.
        /// Only the first matching handler will be removed, if multiple handlers have been added.
        /// </summary>        
        public static void RemoveHandler(INotifyPropertyChanged objectToWatch, AccumulatedPropertyChangedHandler handler)
        {
            lock (timer)
            {


                var toRemove = notifications.FirstOrDefault(n =>
                    Object.ReferenceEquals(n.Item1, objectToWatch) &&
                    n.Item2.Equals(handler));

                if (toRemove != null)
                    notifications.Remove(toRemove);

                // If no longer have any notifications for this object, unhand the property change indicator
                if (!AreWatchingObject(objectToWatch))
                {
                    objectToWatch.PropertyChanged -= objectToWatch_PropertyChanged;
                }
            }
        }


        private static void objectToWatch_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            lock (timer)
            {
                var obj = (INotifyPropertyChanged)sender;
                if (!AreWatchingObject(obj))
                    return; // ignore

                changeNotificationsQueued.Add(new Tuple<INotifyPropertyChanged, PropertyChangedEventArgs, DateTime>(obj, e, DateTime.Now));
            }
        }
    }

    public class AccumulatedPropertyChangedEventArgs : EventArgs 
    {
        /// <summary>
        /// A distinct list of properties changed within the last notification cycle. If any notification
        /// of ""/null (meaning possibly all properties) has been submitted, then this property will be null.
        /// </summary>
        public IEnumerable<string> Properties {get;set;}
    }
}
