﻿using System;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Wsbis.CustomControls;
using System.ComponentModel;
using System.Collections.Generic;
using System.Linq;

namespace UnitTestProject1
{
    [TestClass]
    public class AccumulatedPropertyChangeNotificationTest
    {

        public class TestClass : INotifyPropertyChanged
        {
            private string _a;
            public string A
            {
                get
                {
                    return _a;
                }
                set
                {
                    _a = value;
                    if (PropertyChanged != null)
                        PropertyChanged(this, new PropertyChangedEventArgs("A"));
                }
            }

            private string _b;
            public string B
            {
                get
                {
                    return _b;
                }
                set
                {
                    _b = value;
                    if (PropertyChanged != null)
                        PropertyChanged(this, new PropertyChangedEventArgs("B")); 
                }
            }

            public void NotifyAll()
            {
                if (PropertyChanged != null)
                    PropertyChanged(this, new PropertyChangedEventArgs("")); 
            }

            public event PropertyChangedEventHandler PropertyChanged;
        }

        List<Tuple<object, AccumulatedPropertyChangedEventArgs>> events = new List<Tuple<object, AccumulatedPropertyChangedEventArgs>>();

        public void Handler(object sender, AccumulatedPropertyChangedEventArgs e)
        {
            events.Add(new Tuple<object, AccumulatedPropertyChangedEventArgs>(sender, e));
        }

        [TestMethod]
        public void APCNBasicTest()
        {
            var ca = new TestClass();
            var cb = new TestClass();
            var cc = new TestClass();
            

            int cnt = 0;

            AccumulatedPropertyChangeNotification.AddHandler(ca, Handler);
            AccumulatedPropertyChangeNotification.AddHandler(cb, (s2, e2) => { cnt++; });
            AccumulatedPropertyChangeNotification.AddHandler(cc, Handler);

            // 1. First, test case when only one object has changes. Nothing unusual.
            events.Clear();
            ca.A = "hi";
            ca.B = "ho";
            ca.A = "he";
            Assert.AreEqual(0, events.Count);
            Assert.AreEqual(0, cnt);
            System.Threading.Thread.Sleep(AccumulatedPropertyChangeNotification.NOTIFICATION_DELAY_MS * 3);

            Assert.AreEqual(1, events.Count);
            Assert.AreEqual(0, cnt);

            Assert.AreEqual(ca, events[0].Item1);
            Assert.AreEqual(2, events[0].Item2.Properties.Count());
            Assert.IsTrue(events[0].Item2.Properties.Contains("A"));
            Assert.IsTrue(events[0].Item2.Properties.Contains("B"));          

            
        }

        [TestMethod]
        public void APCNMultiObjBasicTest()
        {
            var ca = new TestClass();
            var cb = new TestClass();
            var cc = new TestClass();


            int cnt = 0;

            AccumulatedPropertyChangeNotification.AddHandler(ca, Handler);
            AccumulatedPropertyChangeNotification.AddHandler(cb, (s2, e2) => { cnt++; });
            AccumulatedPropertyChangeNotification.AddHandler(cc, Handler);

            // Test the case when multiple objects are receiving changes
            events.Clear();
            ca.A = "hi";
            cb.A = "yo";
            ca.B = "ho";
            cc.NotifyAll();
            ca.A = "he";
            Assert.AreEqual(0, events.Count);
            Assert.AreEqual(0, cnt);
            System.Threading.Thread.Sleep(AccumulatedPropertyChangeNotification.NOTIFICATION_DELAY_MS * 3);

            Assert.AreEqual(2, events.Count);
            Assert.AreEqual(1, cnt);

            Assert.IsTrue(ca == events[0].Item1 || ca == events[1].Item1);
            Assert.IsTrue(cc == events[0].Item1 || cc == events[1].Item1);

            var cae = events[0];
            var cce = events[1];
            if (Object.ReferenceEquals(ca, events[1]))
            {
                cae = events[1];
                cce = events[0];
            }
            
            Assert.AreEqual(2, cae.Item2.Properties.Count());
            Assert.IsTrue(cae.Item2.Properties.Contains("A"));
            Assert.IsTrue(cae.Item2.Properties.Contains("B"));

            Assert.IsNull(cce.Item2.Properties); // should be null because of notify-all


        }

        [TestMethod]
        public void APCNObjSequenceTest()
        {
            var ca = new TestClass();
            var cb = new TestClass();
            int cnt = 0;

            AccumulatedPropertyChangeNotification.AddHandler(ca, Handler);
            AccumulatedPropertyChangeNotification.AddHandler(cb, (s2, e2) => { cnt++; });

            // Test the delay capability by sending seq of events
            events.Clear();
            ca.A = "hi";
            System.Threading.Thread.Sleep(5);
            Assert.AreEqual(0, events.Count);
            ca.B = "ho";
            System.Threading.Thread.Sleep(5);
            Assert.AreEqual(0, events.Count);
            ca.A = "hi";
            System.Threading.Thread.Sleep(5);
            Assert.AreEqual(0, events.Count);

            System.Threading.Thread.Sleep(AccumulatedPropertyChangeNotification.NOTIFICATION_DELAY_MS * 3);
            Assert.AreEqual(1, events.Count);
            Assert.AreEqual(2, events[0].Item2.Properties.Count());

            // Test the delay capability by sending events in rapid succession
            events.Clear();
            ca.A = "hi";            
            Assert.AreEqual(0, events.Count);
            Assert.AreEqual(0, cnt);

            System.Threading.Thread.Sleep(5);
            Assert.AreEqual(0, events.Count);
            Assert.AreEqual(0, cnt);


            for (int i = 0; i < 100; i++)
            {
                // the whole time, no notifications should occur
                ca.A = "ho";
                if (i == 20)
                    cb.B = "xx"; // only once, will notify before loop ends
                System.Threading.Thread.Sleep(5);
                Assert.AreEqual(0, events.Count);
            }
            Assert.AreEqual(1, cnt);
            Assert.AreEqual(0, events.Count);
            System.Threading.Thread.Sleep(AccumulatedPropertyChangeNotification.NOTIFICATION_DELAY_MS * 3);
            Assert.AreEqual(1, events.Count);
            Assert.AreEqual(1, cnt);






        }

        [TestMethod]
        public void APCNAllCoalesceNullTest()
        {
            var ca = new TestClass();            

            AccumulatedPropertyChangeNotification.AddHandler(ca, Handler);
            events.Clear();
            ca.A = "hi";
            System.Threading.Thread.Sleep(5);
            ca.NotifyAll();
            System.Threading.Thread.Sleep(5);
            ca.B = "hi";
            System.Threading.Thread.Sleep(AccumulatedPropertyChangeNotification.NOTIFICATION_DELAY_MS * 3);

            Assert.AreEqual(1, events.Count);
            Assert.AreEqual(null, events[0].Item2.Properties);

        }

        [TestMethod]
        public void APCNAddRemoveTest()
        {
            var ca = new TestClass();
            var cb = new TestClass();
            int cnt = 0;
            events.Clear();

            AccumulatedPropertyChangeNotification.AddHandler(ca, Handler);
            AccumulatedPropertyChangeNotification.AddHandler(cb, (s2, e2) => { cnt++; });
            
            
            ca.A = "hi";
            System.Threading.Thread.Sleep(AccumulatedPropertyChangeNotification.NOTIFICATION_DELAY_MS * 3);
            cb.A = "hi";
            System.Threading.Thread.Sleep(AccumulatedPropertyChangeNotification.NOTIFICATION_DELAY_MS * 3);

            Assert.AreEqual(1, events.Count());
            Assert.AreEqual(1, cnt);

            AccumulatedPropertyChangeNotification.RemoveHandler(ca, Handler);

            ca.A = "hi";
            System.Threading.Thread.Sleep(AccumulatedPropertyChangeNotification.NOTIFICATION_DELAY_MS * 3);
            cb.A = "hi";
            System.Threading.Thread.Sleep(AccumulatedPropertyChangeNotification.NOTIFICATION_DELAY_MS * 3);

            Assert.AreEqual(1, events.Count()); // should not increase, not handled anymore
            Assert.AreEqual(2, cnt);



        }


        [TestMethod]
        public void APCNAddRemoveSameObjTest()
        {
            var ca = new TestClass();            
            int cnt = 0;
            events.Clear();

            AccumulatedPropertyChangeNotification.AddHandler(ca, Handler);
            AccumulatedPropertyChangeNotification.AddHandler(ca, (s2, e2) => { cnt++; });


            ca.A = "hi";
            System.Threading.Thread.Sleep(AccumulatedPropertyChangeNotification.NOTIFICATION_DELAY_MS * 3);
            Assert.AreEqual(1, events.Count());
            Assert.AreEqual(1, cnt);

            ca.B = "hi";
            System.Threading.Thread.Sleep(AccumulatedPropertyChangeNotification.NOTIFICATION_DELAY_MS * 3);
            Assert.AreEqual(2, events.Count());
            Assert.AreEqual(2, cnt);

            AccumulatedPropertyChangeNotification.RemoveHandler(ca, Handler);

            ca.A = "ho";
            System.Threading.Thread.Sleep(AccumulatedPropertyChangeNotification.NOTIFICATION_DELAY_MS * 3);
            Assert.AreEqual(2, events.Count()); // should not increase, not handled anymore
            Assert.AreEqual(3, cnt);

            ca.B = "ho";
            System.Threading.Thread.Sleep(AccumulatedPropertyChangeNotification.NOTIFICATION_DELAY_MS * 3);
            Assert.AreEqual(2, events.Count()); // should not increase, not handled anymore
            Assert.AreEqual(4, cnt);



        }
    }
}
